#include <stdio.h>
#include "xlsc.h"

typedef struct y8_mux_t y8_mux_t;

y8_mux_t *y8_in_open( char *filename, int width, int height );
int y8_in_get_frame_total( y8_mux_t *h );
int y8_in_read_frame( y8_mux_t *h, xlsc_picture_t *pic, int frame );
void y8_in_close( y8_mux_t *h );

y8_mux_t *y8_out_open( char *filename, int width, int height );
int y8_out_write_frame( y8_mux_t *h, xlsc_picture_t *pic );
void y8_out_close( y8_mux_t *h );

int bsf_out_write_frame( FILE *fh, uint8_t *data, int size );

int bsf_in_read_frame( FILE *fh, uint8_t **data, int *size );
